from jaseci.jsorc.live_actions import jaseci_action
from jaseci.utils.utils import logger
import torch
from transformers import pipeline
from transformers import AutoModelForCausalLM, AutoTokenizer
from peft import prepare_model_for_int8_training, LoraConfig, get_peft_model, PeftModel
from datasets import load_dataset
import transformers
import os

DEFAULT_TRAINING_PARAMS = {
    "MICRO_BATCH_SIZE": 4,
    "BATCH_SIZE": 128,
    "EPOCHS": 2,
    "LEARNING_RATE": 2e-5,
    "CUTOFF_LEN": 256,
    "LORA_R": 4,
    "LORA_ALPHA": 16,
    "LORA_DROPOUT": 0.05,
}


@jaseci_action(act_group=["llm"], allow_remote=True)
def setup(
    model_name: str = "databricks/dolly-v2-3b",
    lora_dir: str = None,
    tokenizer_name: str = None,
    **kwargs,
):
    global pipeline, model, tokenizer

    tokenizer_kwargs = kwargs.get("tokenizer_kwargs", {})
    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name if tokenizer_name else model_name, **tokenizer_kwargs
    )
    model_kwargs = kwargs.get("model_kwargs", {})
    model = AutoModelForCausalLM.from_pretrained(
        model_name,
        device_map="auto",
        torch_dtype=torch.float16,
        load_in_8bit=True,
        **model_kwargs,
    )
    logger.info("Model loaded")
    if lora_dir:
        model = PeftModel.from_pretrained(
            model, lora_dir, device_map="auto", torch_dtype=torch.float16
        )
        logger.info("Lora loaded")

    pipeline = pipeline("text-generation", model=model, tokenizer=tokenizer)
    logger.info("Pipeline created")


@jaseci_action(act_group=["llm"], allow_remote=True)
def generate(prompt: str, **kwargs):
    return pipeline(prompt, **kwargs)


@jaseci_action(act_group=["llm"], allow_remote=True)
def train(
    data_file: str,
    output_dir: str,
    training_params=DEFAULT_TRAINING_PARAMS,
    resume_from_checkpoint=False,
    **kwargs,
):
    global model, tokenizer
    os.makedirs(
        output_dir, exist_ok=True
    )  # create output directory if it doesn't exist

    BATCH_SIZE = training_params.get(
        "BATCH_SIZE", DEFAULT_TRAINING_PARAMS["BATCH_SIZE"]
    )
    MICRO_BATCH_SIZE = training_params.get(
        "MICRO_BATCH_SIZE", DEFAULT_TRAINING_PARAMS["MICRO_BATCH_SIZE"]
    )
    EPOCHS = training_params.get("EPOCHS", DEFAULT_TRAINING_PARAMS["EPOCHS"])
    LEARNING_RATE = training_params.get(
        "LEARNING_RATE", DEFAULT_TRAINING_PARAMS["LEARNING_RATE"]
    )
    CUTOFF_LEN = training_params.get(
        "CUTOFF_LEN", DEFAULT_TRAINING_PARAMS["CUTOFF_LEN"]
    )
    LORA_R = training_params.get("LORA_R", DEFAULT_TRAINING_PARAMS["LORA_R"])
    LORA_ALPHA = training_params.get(
        "LORA_ALPHA", DEFAULT_TRAINING_PARAMS["LORA_ALPHA"]
    )
    LORA_DROPOUT = training_params.get(
        "LORA_DROPOUT", DEFAULT_TRAINING_PARAMS["LORA_DROPOUT"]
    )
    GRADIENT_ACCUMULATION_STEPS = BATCH_SIZE // MICRO_BATCH_SIZE

    # Preparing the model for int8 training
    model = prepare_model_for_int8_training(model, use_gradient_checkpointing=True)
    config = LoraConfig(
        r=LORA_R,
        lora_alpha=LORA_ALPHA,
        lora_dropout=LORA_DROPOUT,
        bias="none",
        task_type="CAUSAL_LM",
        **kwargs.get("lora_config_kwargs", {}),
    )
    model = get_peft_model(model, config)
    tokenizer.pad_token_id = 0

    # Preparing the dataset
    data = load_dataset("json", data_files=data_file)
    data = data.shuffle().map(
        lambda data_point: tokenizer(
            generate_prompt(data_point),
            truncation=True,
            max_length=CUTOFF_LEN,
            padding="max_length",
        )
    )

    trainer = transformers.Trainer(
        model=model,
        train_dataset=data["train"],
        args=transformers.TrainingArguments(
            per_device_train_batch_size=MICRO_BATCH_SIZE,
            gradient_accumulation_steps=GRADIENT_ACCUMULATION_STEPS,
            warmup_steps=20,
            num_train_epochs=EPOCHS,
            learning_rate=LEARNING_RATE,
            fp16=True,
            logging_steps=1,
            output_dir=output_dir,
            save_total_limit=3,
            **kwargs.get("training_args_kwargs", {}),
        ),
        data_collator=transformers.DataCollatorForLanguageModeling(
            tokenizer, mlm=False
        ),
    )
    model.config.use_cache = False

    trainer.train(resume_from_checkpoint=resume_from_checkpoint)

    # Saving the model
    os.makedirs(os.path.join(output_dir, "final"), exist_ok=True)
    model.save_pretrained(os.path.join(output_dir, "final"))


@jaseci_action(act_group=["llm"], allow_remote=True)
def generate_prompt(data_point):
    # taken from https://github.com/tloen/alpaca-lora
    if data_point["input"]:
        return f"""Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.

        ### Instruction:
        {data_point["instruction"]}

        ### Input:
        {data_point["input"]}

        ### Response:
        {data_point["output"]}"""
    else:
        return f"""Below is an instruction that describes a task. Write a response that appropriately completes the request.

        ### Instruction:
        {data_point["instruction"]}

        ### Response:
        {data_point["output"]}"""


@jaseci_action(act_group=["llm"], allow_remote=True)
def run_dataset_builder():
    import streamlit.web.bootstrap

    streamlit.web.bootstrap.run(
        os.path.join(os.path.dirname(__file__), "dataset_builder.py"), "", [], []
    )



FAST_ENC_ACTION_CONFIG = {
    "module": "jac_nlp.fast_enc",
    "loaded_module": "jac_nlp.fast_enc.fast_enc",
    "remote": {
        "Service": {
            "kind": "Service",
            "apiVersion": "v1",
            "metadata": {"name": "fast-enc", "creationTimestamp": None},
            "spec": {
                "ports": [
                    {"name": "http", "protocol": "TCP", "port": 80, "targetPort": 80}
                ],
                "selector": {"pod": "fast-enc"},
                "type": "ClusterIP",
                "sessionAffinity": "None",
                "internalTrafficPolicy": "Cluster",
            },
            "status": {"loadBalancer": {}},
        },
        "ConfigMap": {
            "kind": "ConfigMap",
            "apiVersion": "v1",
            "metadata": {
                "name": "fast-enc-up",
                "creationTimestamp": None,
            },
            "data": {
                "prod_up": "uvicorn jac_nlp.fast_enc:serv_actions --host 0.0.0.0 --port 80"
            },
        },
        "Deployment": {
            "kind": "Deployment",
            "apiVersion": "apps/v1",
            "metadata": {"name": "fast-enc", "creationTimestamp": None},
            "spec": {
                "replicas": 1,
                "selector": {"matchLabels": {"pod": "fast-enc"}},
                "template": {
                    "metadata": {
                        "name": "fast-enc",
                        "creationTimestamp": None,
                        "labels": {"pod": "fast-enc"},
                    },
                    "spec": {
                        "volumes": [
                            {
                                "name": "prod-script",
                                "configMap": {
                                    "name": "fast-enc-up",
                                    "defaultMode": 420,
                                },
                            }
                        ],
                        "containers": [
                            {
                                "name": "fast-enc",
                                "image": "jaseci/jac-nlp:latest",
                                "command": ["bash", "-c", "source script/prod_up"],
                                "ports": [{"containerPort": 80, "protocol": "TCP"}],
                                "resources": {
                                    "limits": {"memory": "3Gi"},
                                    "requests": {"memory": "3Gi"},
                                },
                                "volumeMounts": [
                                    {"name": "prod-script", "mountPath": "/script"}
                                ],
                                "terminationMessagePath": "/dev/termination-log",
                                "terminationMessagePolicy": "File",
                                "imagePullPolicy": "IfNotPresent",
                            }
                        ],
                        "restartPolicy": "Always",
                        "terminationGracePeriodSeconds": 30,
                        "dnsPolicy": "ClusterFirst",
                        "securityContext": {},
                        "schedulerName": "default-scheduler",
                    },
                },
                "strategy": {
                    "type": "RollingUpdate",
                    "rollingUpdate": {"maxUnavailable": "25%", "maxSurge": "25%"},
                },
                "revisionHistoryLimit": 10,
                "progressDeadlineSeconds": 600,
            },
            "status": {},
        },
    },
}



import streamlit as st
import pandas as pd

st.header("LLM Dataset Builder")

# stop re-running
if "dataset" not in st.session_state:
    st.session_state["dataset"] = pd.DataFrame(
        columns=["instruction", "input", "output"]
    )

instruction = st.text_area("Instruction")
_input = st.text_area("Input")
output = st.text_area("Output")

if st.button("Add"):
    if instruction and output:
        st.session_state["dataset"] = pd.concat(
            [
                st.session_state["dataset"],
                pd.DataFrame(
                    [[instruction, _input, output]],
                    columns=["instruction", "input", "output"],
                ),
            ],
            ignore_index=True,
        )
    else:
        st.error("Instruction and output are required")

st.dataframe(st.session_state["dataset"], use_container_width=True)

if st.button("Save"):
    st.session_state["dataset"].to_json("dataset.json", orient="records")



import torch
import os
import configparser
import warnings

from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from jaseci.jsorc.live_actions import jaseci_action
from jaseci.jsorc.remote_actions import launch_server
from fastapi import HTTPException
from traceback import print_exc

warnings.filterwarnings("ignore")
warnings.warn("ignore")

config = configparser.ConfigParser()
config.read(os.path.join(os.path.dirname(__file__), "config.cfg"))

model = AutoModelForSeq2SeqLM.from_pretrained(config["MODEL"]["T5-SMALL"])
tokenizer = AutoTokenizer.from_pretrained(config["TOKENIZER"]["T5-SMALL"])

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model = model.to(device)


@jaseci_action(act_group=["setup"], allow_remote=True)
def setup(model_name: str = "T5-SMALL", tokenizer_name: str = "T5-SMALL"):
    global model, tokenizer

    model = AutoModelForSeq2SeqLM.from_pretrained(config["MODEL"][model_name]).to(
        device
    )
    tokenizer = AutoTokenizer.from_pretrained(config["TOKENIZER"][tokenizer_name])


@jaseci_action(act_group=["paraphraser"], allow_remote=True)
def paraphrase(text: str):
    try:
        encoding = tokenizer.encode_plus(
            text, max_length=128, padding=True, return_tensors="pt"
        )
        input_ids, attention_mask = encoding["input_ids"].to(device), encoding[
            "attention_mask"
        ].to(device)

        diverse_beam_outputs = model.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            max_length=128,
            early_stopping=True,
            num_beams=5,
            num_beam_groups=5,
            num_return_sequences=5,
            diversity_penalty=0.70,
        )
        paraphrase_list = []

        for beam_output in diverse_beam_outputs:
            sent = tokenizer.decode(
                beam_output, skip_special_tokens=True, clean_up_tokenization_spaces=True
            )
            paraphrase_list.append(sent)

        return paraphrase_list

    except Exception as e:
        print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    launch_server(port=8000)



from jaseci.jsorc.live_actions import jaseci_action
import torch
from transformers import pipeline
from langchain import PromptTemplate, LLMChain
from langchain.llms import HuggingFacePipeline


@jaseci_action(act_group=["dolly"], allow_remote=True)
def setup(
    model: str = "dolly-v2-3b", input_variables: list = None, template: str = None
):
    global llm_chain, llm_context_chain, custom_llm_chain, custom_prompt
    dolly_pipeline = pipeline(
        model=f"databricks/{model}",
        torch_dtype=torch.bfloat16,
        trust_remote_code=True,
        device_map="auto",
        return_full_text=True,
    )
    prompt = PromptTemplate(input_variables=["instruction"], template="{instruction}")
    prompt_with_context = PromptTemplate(
        input_variables=["instruction", "subject"],
        template="{instruction}\n\nInput:\n{subject}",
    )

    if input_variables and template:
        custom_prompt = PromptTemplate(
            input_variables=input_variables, template=template
        )
    else:
        custom_prompt = None

    hf_pipeline = HuggingFacePipeline(pipeline=dolly_pipeline)

    llm_chain = LLMChain(llm=hf_pipeline, prompt=prompt)
    llm_context_chain = LLMChain(llm=hf_pipeline, prompt=prompt_with_context)
    if custom_prompt:
        custom_llm_chain = LLMChain(llm=hf_pipeline, prompt=custom_prompt)


@jaseci_action(act_group=["dolly"], allow_remote=True)
def generate(prompt: str = None, subject: str = None, **kwargs):
    global llm_chain, llm_context_chain, custom_llm_chain, custom_prompt
    if custom_prompt:
        return custom_llm_chain.predict(**kwargs).lstrip()
    if subject is None and prompt:
        return llm_chain.predict(instruction=prompt).lstrip()
    elif subject and prompt:
        return llm_context_chain.predict(instruction=prompt, subject=subject).lstrip()
    else:
        raise Exception(
            "No custom prompt specified, and no subject context or prompt given."
        )
